/*
 * Copyright (C) 2023  Valentin Lorentz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use crate::widgets::Prerender;
use futures::stream::FuturesUnordered;
use futures::StreamExt;
use matrix_sdk::async_trait;
use nonempty::NonEmpty;
use ratatui::text::Text;
use std::collections::HashSet;

mod log;
pub use log::LogBuffer;
mod room;
pub use room::RoomBuffer;

#[derive(Clone, Debug, PartialEq, Eq, Hash)]
pub enum BufferId {
  /// The main/home buffer
  Log,
  /// Any Matrix room
  Room(matrix_sdk::ruma::OwnedRoomId),
}

#[derive(Debug, Clone)]
pub enum BufferItemContent<'buf> {
  Text(Text<'buf>),
  Divider(Text<'buf>),
  Empty,
}

pub struct BufferItem<'buf> {
  pub content: BufferItemContent<'buf>,
  pub prerender: &'buf Prerender,
}

#[async_trait]
pub trait Buffer: Send + Sync + memuse::DynamicUsage {
  /// A short human-readable name for the room, eg. to show in compact buflist
  fn short_name(&self) -> String;
  /// Unique identifier of this buffer
  fn id(&self) -> BufferId;
  /// Identifier of this buffer's parent buffer, if any.
  fn parent(&self) -> Option<BufferId> {
    None
  }
  /// Returns if there are any updates to apply.
  async fn poll_updates(&mut self);
  fn content<'a>(&'a self) -> Box<dyn Iterator<Item = BufferItem<'a>> + 'a>;
  /// Called when the user is being showned the oldest items this buffer returned.
  ///
  /// This should return immediately, not waiting for anything to be loaded.
  fn request_back_pagination(&self, num: u16) {}

  fn dynamic_usage(&self) -> usize {
    memuse::DynamicUsage::dynamic_usage(self)
  }
  fn dynamic_usage_bounds(&self) -> (usize, Option<usize>) {
    memuse::DynamicUsage::dynamic_usage_bounds(self)
  }
}

pub struct Buffers {
  buffers: Vec<Box<dyn Buffer>>,
  active_index: usize,
}

impl Buffers {
  pub fn new(initial_buffer: Box<dyn Buffer>) -> Self {
    Self {
      buffers: vec![initial_buffer],
      active_index: 0,
    }
  }

  pub async fn poll_updates(&mut self) {
    self
      .iter_mut()
      .map(|buf| buf.poll_updates())
      .collect::<FuturesUnordered<_>>()
      .next()
      .await
      .expect("poll_updates reached the end of the never-ending stream");

    // Reorder buffers in case we just got an update on space relationships
    // FIXME: do this only when needed
    let mut buffers = Vec::new();
    std::mem::swap(&mut self.buffers, &mut buffers);
    for buf in buffers.into_iter() {
      self.push(buf);
    }
  }

  pub fn len(&self) -> usize {
    self.buffers.len()
  }

  pub fn iter(&self) -> impl Iterator<Item = &dyn Buffer> {
    self.buffers.iter().map(|buffer_box| &**buffer_box)
  }

  pub fn iter_mut(&mut self) -> impl Iterator<Item = &mut Box<dyn Buffer>> {
    self.buffers.iter_mut()
  }

  /// `O(self.len())` if the buffer has a parent
  pub fn push(&mut self, new_buffer: Box<dyn Buffer>) {
    let new_buffer_id = new_buffer.id();

    match new_buffer
      .parent()
      .and_then(|parent_id| self.buffers.iter().position(|buf| buf.id() == parent_id))
    {
      None => self.buffers.push(new_buffer),
      Some(parent_position) => {
        // iterator through buffers after the parent, and record on the stack the chain
        // of successor from the parent to the current buffer. As soon as we see a buffer
        // with a parent not on the stack (or with no parent), it means we left the subtree,
        // and should insert the new buffer there.
        let mut stack = vec![self.buffers[parent_position].id()];
        let mut new_buffer = Some(new_buffer);
        for (i, buf) in self.buffers.iter().enumerate().skip(parent_position + 1) {
          match buf.parent() {
            None => {}, // root buffer
            Some(parent_id) => {
              match stack.iter().position(|id| *id == parent_id) {
                None => {}, // parent is not in the subtree
                Some(parent_position) => {
                  // parent is in the subtree
                  stack.truncate(parent_position + 1);
                  stack.push(buf.id());
                  continue;
                },
              }
            },
          }
          // if we are here, it means buf.parent() is not in the subtree (or buf is orphan)
          // so we should insert the new buffer at this position
          self
            .buffers
            .insert(i, new_buffer.take().expect("new_buffer was already taken"));
          break;
        }

        if let Some(new_buffer) = new_buffer {
          // The parent's subtree was the last subtree
          self.buffers.push(new_buffer);
        }
      },
    }

    // Find existing children and move them after this buffer (while preserving their
    // relative order.

    // 1. collect the set of children
    let mut children_ids = HashSet::new();
    children_ids.insert(new_buffer_id.clone());
    let mut added_children_ids = true;
    while added_children_ids {
      added_children_ids = false;
      for buf in self.buffers.iter() {
        if children_ids.contains(&buf.id()) {
          // Already collected
          continue;
        }
        if let Some(parent) = buf.parent() {
          if children_ids.contains(&parent) {
            children_ids.insert(buf.id());
          }
        }
      }
    }
    children_ids.remove(&new_buffer_id);

    // 2. pop them from the list of buffers
    let mut all_buffers = Vec::new();
    let mut child_buffers = Vec::new();
    std::mem::swap(&mut all_buffers, &mut self.buffers);
    for buf in all_buffers.into_iter() {
      if children_ids.contains(&buf.id()) {
        child_buffers.push(buf)
      } else {
        self.buffers.push(buf);
      }
    }

    // 3. split the set of buffers, and insert the children after the new buffer
    let after_children = self.buffers.split_off(
      self
        .buffers
        .iter()
        .position(|buf| buf.id() == new_buffer_id)
        .expect("new buffer disappeared")
        + 1,
    );
    self.buffers.extend(child_buffers.into_iter());
    self.buffers.extend(after_children);
  }

  pub fn active_index(&self) -> usize {
    self.active_index
  }

  pub fn set_active_index(&mut self, index: isize) {
    self.active_index = index.wrapping_rem_euclid(self.buffers.len() as isize) as usize;
  }

  pub fn active_buffer(&self) -> &dyn Buffer {
    &**self
      .buffers
      .get(self.active_index)
      .expect("Active buffer index does not exist")
  }

  pub fn active_buffer_mut(&mut self) -> &mut Box<dyn Buffer> {
    self
      .buffers
      .get_mut(self.active_index)
      .expect("Active buffer index does not exist")
  }
}
